# Copyright (c) 2006 Charl P. Botha <http://cpbotha.net/>
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from ee_except import EnvEditException, EnvEditWarningException
import os
from Tkinter import *
import tkMessageBox
import tkFileDialog
from widget_redirector import WidgetRedirector

MODE_SYNCED=0
MODE_MODIFIED=1
MODE_PREVIEW=2

MODE_TO_TEXT = {MODE_SYNCED : "= SYNCHRONISED =",
                MODE_MODIFIED : "* MODIFIED *",
                MODE_PREVIEW : "- PREVIEW -"}

CURTEXT_USER=0
CURTEXT_SYS=1

class StatusBar(Frame):
    """From the tkinterbook.
    """

    def __init__(self, master):
        Frame.__init__(self, master)
        self.label = Label(self, bd=1, relief=SUNKEN, anchor=W)
        self.label.pack(fill=X)

    def set(self, format, *args):
        self.label.config(text=format % args)
        self.label.update_idletasks()

    def clear(self):
        self.label.config(text="")
        self.label.update_idletasks()

class RedirectedText(Text):
    """We would like to know when the Text widget's contents change.  We can't
    just override the insert method, we have to make use of some Tk magic.
    This magic is encapsulated in the idlelib.WidgetRedirector class which
    we use here.
    """

    def __init__(self, master=None, cnf={}, **kw):
        Text.__init__(self, master, cnf, **kw)

        # now attach the redirector
        self.redir = WidgetRedirector(self)
        self.orig_insert = self.redir.register("insert", self.new_insert)

    def new_insert(self, *args):
        self.event_generate('<<Insert>>')
        self.orig_insert(*args)
    

class View:

    def __init__(self, controller):
        self.controller = controller

        self.root = None
        self.user_text = None
        self._user_mode_label_var = None
        self._user_mode = None
        self.sys_text = None
        self._sys_mode_label_var = None
        self._sys_mode = None
        
        self._create_ui()

        self._current_text = None
        self.user_text.focus_set()

    def close(self):
        self.root.destroy()

    def _create_ui(self):

        # these two variables determine the final dimensions of our interface
        HALF_FRAME_HEIGHT=235
        TEXT_WIDTH=100
        
        self.root = Tk()
        self.root.title(
            "envedit %s Environment Editor (c) 2006 by Charl P. Botha" %
            (self.controller.get_version(),))

        # try finding icon in resdir and in appdir
        try:
            self.root.iconbitmap(
                os.path.join(self.controller.get_resdir(), 'envedit.ico'))
            
        except TclError:
            try:
                self.root.iconbitmap(
                    os.path.join(self.controller.get_appdir(), 'envedit.ico'))
                
            except TclError:
                self.show_error('Error', 'Error opening icon.  Continuing.')

        # MAIN MENU ####################################################

        menu = Menu(self.root)
        self.root.config(menu=menu)

        file_menu = Menu(menu, tearoff=False)
        menu.add_cascade(label="File", underline='0', menu=file_menu)

        # FILE ##########################################################
        file_menu.add_command(label="Save to TXT file", underline=0,
                              command=self.cmd_save, accelerator="Ctrl+S")
        self.root.bind_all("<Control-s>", self.cmd_save)
        
        file_menu.add_command(label="Load from TXT file", underline=0,
                              command=self.cmd_load, accelerator="Ctrl+L")
        self.root.bind_all("<Control-l>", self.cmd_load)
        
        file_menu.add_separator()

        # currently disabled, the process needs special privileges
        #file_menu.add_command(label="Export to REG file", underline=0,
        #                      command=self.cmd_export)
        #
        #file_menu.add_separator()

        file_menu.add_command(label = "Exit", underline=1,
                              command=self.cmd_exit, accelerator="Ctrl+Q")
        self.root.bind_all("<Control-q>", self.cmd_exit)

        # ENVIRONMENT ###################################################
        env_menu = Menu(menu, tearoff=False)
        menu.add_cascade(label="Environment", underline=0, menu=env_menu)

        env_menu.add_command(label = "Sync with environment", underline=1,
                              command = self.cmd_sync_with_environment,
                             accelerator="F5")
        self.root.bind_all("<F5>", self.cmd_sync_with_environment)
        
        env_menu.add_command(label = "Preview changes to environment",
                             underline=0, command=self.cmd_preview,
                             accelerator="F7")
        self.root.bind_all("<F7>", self.cmd_preview)
        
        env_menu.add_command(label = "Apply to environment", underline = 0,
                              command = self.cmd_apply_to_environment,
                             accelerator="F8")
        self.root.bind_all("<F8>", self.cmd_apply_to_environment)
        
        

        # HELP ##########################################################
        help_menu = Menu(menu, tearoff=False)
        menu.add_cascade(label="Help", underline='0', menu=help_menu)

        help_menu.add_command(label = "About", underline = 0,
                              command = self.cmd_help_about)

        # END MENU ######################################################
        
        separator = Frame(self.root, height=2, bd=1, relief=SUNKEN)
        separator.pack(fill=X, padx=5, pady=2)


        paned_window = PanedWindow(self.root, orient=VERTICAL)
        paned_window.pack(fill=BOTH, expand=1)
        
        top_frame = Frame(paned_window, height=HALF_FRAME_HEIGHT)
        paned_window.add(top_frame)

        bottom_frame = Frame(paned_window, height=HALF_FRAME_HEIGHT)
        paned_window.add(bottom_frame)

        # we'll use this method to create the different edit boxes
        def create_scrolled_text(master, label_text):

            label_frame = Frame(master)

            l = Label(label_frame, text = label_text,
                       anchor = W)
            l.pack(side=LEFT)

            mode_label_var = StringVar()
            mode_label_var.set("")
            ml = Label(label_frame, textvariable = mode_label_var, anchor = W)
            ml.pack(side=LEFT)

            label_frame.pack()

            yscrollbar = Scrollbar(master)
            yscrollbar.pack(side=RIGHT, fill=Y)

            # this determines the width of the complete interface (yes)
            text = RedirectedText(master, height=15, width=TEXT_WIDTH,
                                  wrap=CHAR,
                                  yscrollcommand=yscrollbar.set)

            text.pack(fill=BOTH, expand=1)

            #xscrollbar.config(command=text.xview)
            yscrollbar.config(command=text.yview)

            return text, mode_label_var


        # setup user_text ###############################################
        self.user_text, self._user_mode_label_var = \
                        create_scrolled_text(top_frame, "User Environment")

        def cb_ut_fi(event):
            self.set_current_text(CURTEXT_USER)

        self.user_text.bind('<FocusIn>', cb_ut_fi)

        def cb_ut_c(event):
            self.set_user_mode(MODE_MODIFIED)
            
        self.user_text.bind('<<Insert>>', cb_ut_c)

        # setup sys_text ################################################
        self.sys_text, self._sys_mode_label_var = \
                      create_scrolled_text(bottom_frame, "System Environment")

        def cb_st_fi(event):
            self.set_current_text(CURTEXT_SYS)

        self.sys_text.bind('<FocusIn>', cb_st_fi)

        def cb_st_c(event):
            self.set_sys_mode(MODE_MODIFIED)
            
        self.sys_text.bind('<<Insert>>', cb_st_c)

        # setup statusbar ##############################################
        self.statusbar = StatusBar(self.root)
        self.statusbar.set('%s', 'Welcome to envedit!')
        self.statusbar.pack(side=BOTTOM, fill=X)

    def cmd_apply_to_environment(self, event=None):
        if self._current_text == CURTEXT_USER:
            try:
                self.controller.apply_user_env()
                
            except EnvEditException, e:
                tkMessageBox.showerror(
                    'Error',
                    'Could not apply changes to the USER environment: %s' %
                    (str(e),))
                
            except EnvEditWarningException, e:
                tkMessageBox.showwarning(
                    'Warning', str(e))

            else:
                self.statusbar.set(
                    'Successfully applied changes to USER environment.')
                
        else:
            try:
                self.controller.apply_sys_env()
                
            except EnvEditException, e:
                tkMessageBox.showerror(
                    'Error',
                    'Could not apply changes to the USER environment: %s' %
                    (str(e),))
                
            except EnvEditWarningException, e:
                tkMessageBox.showwarning(
                    'Warning', str(e))

            else:
                self.statusbar.set(
                    'Successfully applied changes to SYSTEM environment.')
        

    def cmd_help_about(self):
        tkMessageBox.showinfo(
            'Help | About',
            'envedit %s is copyright 2006 by Charl P. Botha '
            '<http://cpbotha.net/>\n\n'
            'Because the Windows environment editor will '
            'ruin your day, every time.\n\n'
            'That being said, YOU ARE USING ENVEDIT '
            'ENTIRELY *AT YOUR OWN RISK*.' % (self.controller.get_version(),),
            parent = self.root)

    def cmd_load(self, event=None):
        """We handle the complete load request here in the view, as the
        processing is quite trivial and only affects the view.
        """
        
        current_text = self.get_current_text_as_string()
        filename = tkFileDialog.askopenfilename(
            parent=self.root, 
            filetypes=[('Text files', '*.txt'), ('All files', '*')],
            title='Load file into %s environment view' % (current_text,))

        if len(filename) > 0:

            try:
                tf = file(filename, 'r')
                txt = tf.read()
                tf.close()
            except IOError, e:
                tkMessageBox.showerror(
                    'File load error',
                    'Error reading from %s: %s' % (filename, str(e)))
            else:
                if self.get_current_text() == CURTEXT_USER:
                    self.set_user_text(txt)
                    self.set_user_mode(MODE_MODIFIED)

                else:
                    self.set_sys_text(txt)
                    self.set_sys_mode(MODE_MODIFIED)

                self.statusbar.set(
                    'Successfully loaded %s environment view from %s',
                    current_text, filename)

    def cmd_preview(self, event=None):
        if self.get_current_text() == CURTEXT_USER:
            self.controller.preview_user_env()

        else:
            self.controller.preview_sys_env()

    def cmd_save(self, event=None):
        """We handle the complete save request here in the view, as the
        processing is quite trivial and only affects the view.
        """

        current_text = self.get_current_text_as_string()
        filename = tkFileDialog.asksaveasfilename(
            parent=self.root,
            filetypes=[('Text files', '*.txt'), ('All files', '*')],
            title="Save %s environment view to file ..." % (current_text,))
        
        if len(filename) > 0:
            if self.get_current_text() == CURTEXT_USER:
                txt = self.get_user_text()

            else:
                txt = self.get_sys_text()

            try:
                tf = file(filename, 'w')
                tf.write(txt)
                tf.close()
            except IOError, e:
                tkMessageBox.showerror(
                    'File save error',
                    'Error writing to %s: %s' % (filename, str(e)))

            else:
                self.statusbar.set(
                    'Successfully wrote %s environment view to %s.',
                    current_text, filename)

    def cmd_sync_with_environment(self, event=None):
        if self._current_text == CURTEXT_USER:
            try:
                self.controller.sync_user_env()
            except EnvEditException, e:
                self.show_error(
                    'Error',
                    'Could not synchronise with USER environment: %s' %
                    (str(e),))
                
        else:
            try:
                self.controller.sync_sys_env()
            except EnvEditException, e:
                self.show_error(
                    'Error',
                    'Could not synchronise with SYSTEM environment: %s' %
                    (str(e),))
            

    def cmd_exit(self, event=None):
        self.controller.quit()

    def cmd_export(self, event=None):
        """Currently unused, the process needs special privileges for this
        to work.
        """
        
        current_text = self.get_current_text_as_string()
        filename = tkFileDialog.asksaveasfilename(
            parent=self.root,
            filetypes=[('REG files', '*.reg'), ('All files', '*')],
            title="Save %s environment view to REG file ..." %
            (current_text,))
        
        if len(filename) > 0:
            try:
                if self._current_text == CURTEXT_USER:
                    self.controller.export_user_to_reg_file(filename)
                else:
                    self.controller.export_sys_to_reg_file(filename)

            except EnvEditException, e:
                self.show_error(
                    'Error',
                    'Could not export %s environment to REG file: %s' %
                    (current_text, str(e)))

    def main_loop(self):
        self.root.mainloop()

    def get_current_text(self):
        return self._current_text

    def get_current_text_as_string(self):
        if self.get_current_text() == CURTEXT_USER:
            return 'USER'
        else:
            return 'SYSTEM'

    def set_current_text(self, which):
        self._current_text = which
        if self._current_text == CURTEXT_USER:
            ut_col = 'lightblue'
            st_col = 'white'

        else:
            ut_col = 'white'
            st_col = 'lightblue'

        self.user_text.config(bg=ut_col)
        self.sys_text.config(bg=st_col)

    def get_user_text(self):
        """Get and return all text in user edit box.
        """

        return self.user_text.get(1.0, END)

    def get_sys_text(self):
        """Get and return all text in sys edit box.
        """

        return self.sys_text.get(1.0, END)

    def set_user_text(self, txt):
        """Replace the contents of the user_text with txt.
        """
        
        self.user_text.delete(1.0, END) # clear all
        self.user_text.insert(END, txt, "m")

    def set_sys_text(self, txt):
        """Replace the contents of the sys_text with txt.
        """
        
        self.sys_text.delete(1.0, END)
        self.sys_text.insert(END, txt, "m")

    def get_user_mode(self):
        return self._user_mode

    def set_user_mode(self, mode):
        """
        """
        if mode != self._user_mode:
            self._user_mode = mode
            self._user_mode_label_var.set(MODE_TO_TEXT[self._user_mode])

    def set_sys_mode(self, mode):
        if mode != self._sys_mode:
            self._sys_mode = mode
            self._sys_mode_label_var.set(MODE_TO_TEXT[self._sys_mode])

    def get_sys_mode(self):
        return self._sys_mode

    def show_error(self, title, msg):
        tkMessageBox.showerror(title, msg)


